#!/usr/bin/env python3

import numpy as np
from difflib import SequenceMatcher
from collections import OrderedDict, Counter

from .utilities import *
from .variant import Variant, NullVariant
from .localn import findall_mismatches


def phase_nearby_variants(
    target,
    contig,
    pileup,
    basequalthresh,
    snv_neighborhood,
    indel_neighborhood,
    indel_repeat_thresh,
    mut_frac_thresh,
    hard,
    to_complex,
):
    # no indel reads or contig construction failure
    if contig.failed:
        return NullVariant(target.chrom, target.pos, target.reference)

    indexed_contig = contig.contig_dict
    target_pos_on_contig = contig.lt_end_pos

    # no phasable variants
    variants_to_phase = contig.mismatches + contig.non_target_indels
    if not variants_to_phase:
        return make_target_obj_from_contig(target, indexed_contig)

    # phase all phasables within the target exon (hard phasing)
    if hard:
        variants_list = []
        cleaned, variant_list = precleaning(
            indexed_contig, variants_list, target_pos_on_contig, pileup, target
        )
        return greedy_phasing(target, cleaned)
    else:
        indexed_contig, variants_to_phase = precleaning(
            indexed_contig, variants_to_phase, target_pos_on_contig, pileup
        )

    if not variants_to_phase:
        return make_target_obj_from_contig(target, indexed_contig)
    else:
        variants_in_non_targets, mut_frac = variants_in_non_target_pileup(
            pileup, target, basequalthresh, to_complex
        )
        if mut_frac > mut_frac_thresh:
            return make_target_obj_from_contig(target, indexed_contig)

    lt_loci, rt_loci, tmp = [], [], variants_to_phase.copy()
    for var in tmp:
        if is_deletable(var, variants_in_non_targets, indel_repeat_thresh, to_complex):
            if var.pos < target_pos_on_contig:
                lt_loci.append(var.pos)
            elif var.pos > target_pos_on_contig:
                rt_loci.append(var.pos)

            variants_to_phase.remove(var)

    if not variants_to_phase:
        return make_target_obj_from_contig(target, indexed_contig)

    lt_end = max(lt_loci) if lt_loci else -np.inf
    rt_end = min(rt_loci) if rt_loci else np.inf

    remove_deletables(indexed_contig, lt_end, target_pos_on_contig, rt_end)

    mismatches_to_phase = [
        var
        for var in variants_to_phase
        if not var.is_indel and indexed_contig.get(var.pos, False)
    ]
    non_target_indels_to_phase = [
        var
        for var in variants_to_phase
        if var.is_indel and indexed_contig.get(var.pos, False) and var != target
    ]

    if variants_to_phase:
        if not non_target_indels_to_phase:
            peak_locs = locate_mismatch_cluster_peaks(
                indexed_contig,
                mismatches_to_phase,
                target,
                snv_neighborhood,
                to_complex,
            )

            if peak_locs:
                remove_deletables(
                    indexed_contig, peak_locs[0], target_pos_on_contig, peak_locs[1]
                )
            else:
                return make_target_obj_from_contig(target, indexed_contig)
        else:
            target_len = len(target.indel_seq)
            non_target_max_len = max(
                [len(var.indel_seq) for var in non_target_indels_to_phase]
            )

            if max(target_len, non_target_max_len) < 4:
                indel_neighborhood = int(indel_neighborhood / 2) + 1

            remove_common_substrings(
                indexed_contig, target_pos_on_contig, indel_neighborhood
            )

            lt_end = end_point(
                indexed_contig, mismatches_to_phase, target, snv_neighborhood, left=True
            )
            rt_end = end_point(
                indexed_contig,
                mismatches_to_phase,
                target,
                snv_neighborhood,
                left=False,
            )

            remove_deletables(indexed_contig, lt_end, target_pos_on_contig, rt_end)

    cvar = greedy_phasing(target, indexed_contig)

    if cvar != target:
        return cvar
    else:
        return make_target_obj_from_contig(target, indexed_contig)


def make_target_obj_from_contig(target, indexed_contig):
    try:
        data = indexed_contig[target.pos]
        return Variant(
            target.chrom, target.pos, data[0], data[1], target.reference
        ).normalize()
    except:
        return target.normalize()


def greedy_phasing(target, indexed_contig):
    cpos = 0
    cref = ""
    calt = ""
    for k, v in indexed_contig.items():
        if not cpos:
            cpos = k

        cref += v[0]
        calt += v[1]

    return Variant(target.chrom, cpos, cref, calt, target.reference).normalize()


def seq_complexity(contig, snv_neighborhood, indel_neighorhood):
    splits = contig.get_reference_seq(split=True)
    lt_flank, rt_flank = splits[0], splits[2]
    neighorbood = min(snv_neighborhood, indel_neighorhood, len(lt_flank), len(rt_flank))

    return min(
        linguistic_complexity(lt_flank[-neighorbood:]),
        linguistic_complexity(rt_flank[:neighorbood]),
    )


def precleaning(
    genome_indexed_contig, variants_list, target_pos, pileup, limit_to_target_exon=True
):
    lt_loci, rt_loci = [], []

    # filter low qual loci
    for k, v in genome_indexed_contig.items():
        ref, alt, score, cov = v[0], v[1], v[2], v[3]
        if not ref or not alt:
            if k < target_pos:
                lt_loci.append(k)
            elif k > target_pos:
                rt_loci.append(k)

        elif "N" in ref or "N" in alt:
            if k < target_pos:
                lt_loci.append(k)
            elif k > target_pos:
                rt_loci.append(k)

        elif score < score_thresh(ref, alt, cov):
            if k < target_pos:
                lt_loci.append(k)
            elif k > target_pos:
                rt_loci.append(k)

    lt_lim = max(lt_loci) if lt_loci else -np.inf
    rt_lim = min(rt_loci) if rt_loci else np.inf

    if limit_to_target_exon:
        # within the same exon
        spliced_subreads = [
            read["covering_subread"]
            for read in pileup
            if read["is_target"] and read["covering_subread"]
        ]

        if spliced_subreads:
            lt_exon_end = min([subread[0] for subread in spliced_subreads])
            rt_exon_end = max([subread[1] for subread in spliced_subreads])
            lt_lim = max(lt_lim, lt_exon_end - 1)
            rt_lim = min(rt_lim, rt_exon_end + 1)

        tmp = genome_indexed_contig.copy()
        for k, v in genome_indexed_contig.items():
            if k <= lt_lim or rt_lim <= k:
                del tmp[k]

    variants_list = [var for var in variants_list if lt_lim < var.pos < rt_lim]

    return tmp, variants_list


def score_thresh(ref, alt, cov):
    if len(ref) == len(alt) == 1:
        if ref == alt:
            return 0.0
        else:
            if cov > 4:
                return 0.7 if ref == alt else 0.79
            elif 2 < cov <= 4:
                return 0.65
            else:
                return 1.0
    elif len(ref) > 6 or len(alt) > 6:
        return 0.6
    else:
        return 0.67


def locate_mismatch_cluster_peaks(
    indexed_contig, mismatches_to_phase, target, snv_neighborhood, to_complex
):
    lt_peak, lt_peak_pos = calc_peak(
        indexed_contig, mismatches_to_phase, target, snv_neighborhood, left=True
    )
    rt_peak, rt_peak_pos = calc_peak(
        indexed_contig, mismatches_to_phase, target, snv_neighborhood, left=False
    )

    if lt_peak > 0:
        if rt_peak > 0 or rt_peak_pos == np.inf:
            pass
        else:
            return None
    elif rt_peak > 0:
        if lt_peak > 0 or lt_peak_pos == -np.inf:
            pass
        else:
            return None
    else:
        return None

    lt_peak_pos = target.pos if lt_peak_pos == -np.inf else lt_peak_pos
    rt_peak_pos = (
        target.pos + len(target.ref) - 1 if rt_peak_pos == np.inf else rt_peak_pos
    )

    return (lt_peak_pos - 1, rt_peak_pos + 1)


def calc_peak(indexed_contig, mismatches, target, snv_neighborhood, left):
    target_pos = target.pos

    if left:
        loci = [k for k, v in indexed_contig.items() if k <= target_pos][::-1]
        snv_loci = [var.pos for var in mismatches if var.pos < target_pos]
    else:
        del_adjust = len(target.ref) - 1
        loci = [k for k, v in indexed_contig.items() if k > target_pos + del_adjust]
        snv_loci = [var.pos for var in mismatches if var.pos > target_pos]

    score, gain = 0.0, 1.0
    peak_locus = -np.inf if left else np.inf

    if not snv_loci or not loci:
        return score, peak_locus

    indel_len = len(target.indel_seq)
    scores = []
    for i, locus in enumerate(loci):
        if locus in snv_loci:
            score += gain
        else:
            score += loss(i, indel_len, snv_neighborhood)

        scores.append(score)

    peak_score = max(scores)
    if peak_score > 0.0:
        peak_idx = [i for i, j in enumerate(scores) if j == peak_score][-1]
        peak_locus = loci[peak_idx]
        score = peak_score

    return score, peak_locus


def loss(i, indel_len, snv_neighborhood):
    if indel_len < 10:
        return -1 * min(i * 1 / snv_neighborhood, 1.0)
    else:
        return -1 * min(i * 1 * 0.6 / snv_neighborhood, 1.0)


def is_tight_cluster(mismatches, target, snv_neighborhood):
    neigborhood = snv_neighborhood / 2

    lt_near_snvs = [
        var for var in mismatches if target.pos - neigborhood <= var.pos < target.pos
    ]
    lt_far_snvs = [var for var in mismatches if var.pos < target.pos - neigborhood]

    rt_margin = 0 if target.is_ins else len(target.indel_seq)
    rt_near_snvs = [
        var
        for var in mismatches
        if target.pos < var.pos <= target.pos + rt_margin + neigborhood
    ]
    rt_far_snvs = [
        var for var in mismatches if target.pos + rt_margin + neigborhood < var.pos
    ]

    if len(lt_near_snvs) < len(lt_far_snvs):
        return False

    if len(rt_near_snvs) < len(rt_far_snvs):
        return False

    return True


def variants_in_non_target_pileup(pileup, target, basequalthresh, to_complex):
    if not to_complex:
        return [], 0.0

    nontarget_pileup = [
        findall_mismatches(read, end_trim=10)
        for read in pileup
        if not read["is_target"] and read["is_covering"] and not read["is_dirty"]
    ]

    if not nontarget_pileup:
        return [], 0.0

    margin = max(10, min(20, len(target.indel_seq) * 2))
    indels = [
        v[-1]
        for read in nontarget_pileup
        for v in read["I"] + read["D"]
        if not "S" in read["cigar_string"]
        and read["covering_subread"]
        and read["covering_subread"][0] + margin
        < target.pos
        < read["covering_subread"][1] - margin
    ]

    indels = [
        indel
        for indel, cnt in Counter(indels).items()
        if (cnt > 2 and cnt / len(nontarget_pileup) > 0.15) or cnt > 5
    ]

    mismatches = [
        Variant(target.chrom, v[0], v[1], v[2], target.reference)
        for read in nontarget_pileup
        for v in read["mismatches"]
        if v[3] > basequalthresh
    ]

    nontarget_pileup_vol = (
        sum(max(0, len(read["ref_seq"]) - 20) for read in nontarget_pileup) + 1
    )

    mutation_frac = (len(mismatches) + len(indels)) / nontarget_pileup_vol

    mismatches = [
        var
        for var, cnt in Counter(mismatches).items()
        if (cnt > 2 and cnt / len(nontarget_pileup) > 0.15) or cnt > 5
    ]

    return set(indels + mismatches), mutation_frac


def is_deletable(variant, deletable_variants, indel_repeat_thresh, to_complex):
    if to_complex:
        if variant in deletable_variants:
            return True

    if variant.is_indel:
        if repeats(variant) >= indel_repeat_thresh:
            return True

    return False


def repeats(indel):
    unit = to_minimal_repeat_unit(indel.indel_seq)
    return repeat_counter(unit, indel.right_flank())  # leftaligned


def get_freq(freqinfo):
    try:
        wildtype_freq = float(freqinfo.split(",")[0])
    except:
        wildtype_freq = 1.0

    return wildtype_freq


def remove_deletables(indexed_contig, lt_end, target_pos, rt_end):
    tmp = indexed_contig.copy()

    # if lt_end == -np.inf:
    #    lt_end = target_pos - 1
    # if rt_end == np.inf:
    #    rt_end = target_pos + 1

    for k, v in tmp.items():
        if k <= lt_end < target_pos:
            del indexed_contig[k]
        elif lt_end < k < target_pos:
            if v[0] == v[1]:
                del indexed_contig[k]
            else:
                break

    tmp = OrderedDict(reversed(list(tmp.items())))
    for k, v in tmp.items():
        if target_pos < rt_end <= k:
            del indexed_contig[k]
        elif target_pos < k < rt_end:
            if v[0] == v[1]:
                del indexed_contig[k]
            else:
                break

    return indexed_contig


def remove_common_substrings(indexed_contig, target_pos, max_common_str_len):
    common_sub_strs = profile_common_substrings(indexed_contig)

    lt_commons = [sub_str for sub_str in common_sub_strs if sub_str[1] < target_pos]
    rt_commons = [sub_str for sub_str in common_sub_strs if target_pos < sub_str[0]]

    trim_common(indexed_contig, lt_commons, max_common_str_len, left=True)
    trim_common(indexed_contig, rt_commons, max_common_str_len, left=False)

    return indexed_contig


def trim_common(indexed_contig, commons, max_common_str_len, left):
    if not left:
        commons[::-1]

    deletable_commons = []
    for sub_str in commons:
        if sub_str[0] == sub_str[-1]:
            start = sub_str[0]
        else:
            start = search_nearest_lt_locus(indexed_contig, sub_str[0], left)

        end = sub_str[-1]
        start_event = indexed_contig[start]
        end_event = indexed_contig[end]

        # event = start_event if left else end_event

        sub_str_len = end - start
        if sub_str_len >= max_common_str_len:
            if left:
                deletable_commons.append(end)
            else:
                deletable_commons.append(start)

    if deletable_commons:
        loci = [item[0] for item in list(indexed_contig.items())]
        if left:
            lim = max(deletable_commons)
            for locus in loci:
                if locus < lim:
                    del indexed_contig[locus]
        else:
            lim = min(deletable_commons)
            for locus in loci:
                if locus > lim:
                    del indexed_contig[locus]


def search_nearest_lt_locus(indexed_contig, pos, left=True):
    if left:
        not_found = True
    else:
        not_found = False if indexed_contig.get(pos, None) else True

    while not_found:
        pos -= 1

        if indexed_contig.get(pos, False):
            not_found = False
            alleles = indexed_contig[pos]
            ref, alt = alleles[0], alleles[1]

            # when deletion is involved
            if len(ref) > 1:
                pos += len(ref)

    return pos


def profile_common_substrings(indexed_contig):
    commons = []

    items = list(indexed_contig.items())

    contig_pos = items[0][0]
    contig_end = items[-1][0]

    while contig_pos < contig_end:
        common_sub_str = extend_sub_str(contig_pos, indexed_contig)
        end = common_sub_str[-1]
        commons.append(common_sub_str)
        contig_pos = find_next_rt_locus(indexed_contig, end, contig_end)

    return commons


def find_next_rt_locus(indexed_contig, pos, contig_end):
    found = False

    while not found and pos < contig_end:
        pos += 1
        found = indexed_contig.get(pos, False)

    return pos


def extend_sub_str(start, indexed_contig):
    common_start, common_end = start, start

    common_sub_str = []
    for k, v in indexed_contig.items():
        if k > start and v[0] == v[1]:
            common_start = k
            common_sub_str.append(k)
        elif k > common_start > start and v[0] != v[1]:
            common_end = k
            common_sub_str.append(k)
            break

    if not common_sub_str:
        common_sub_str = [common_start, common_end]

    return common_sub_str


def end_point(indexed_contig, mismatches, target, snv_neighborhood, left):
    tmp = indexed_contig.copy()

    if not left:
        tmp = OrderedDict(reversed(list(tmp.items())))

    end_item = list(tmp.items())[0]
    end_pos, end_variant = end_item[0], end_item[1]
    if len(end_variant[0]) != len(end_variant[1]):
        if left:
            return end_pos - 1
        else:
            return end_pos + 1

    end_most_indel = get_end_most_indel(tmp, target)
    if not left:
        tmp = OrderedDict(reversed(list(tmp.items())))

    if not end_most_indel:
        end_most_indel = target

    score, peak_pos = calc_peak(tmp, mismatches, end_most_indel, snv_neighborhood, left)
    if score <= 0:
        if left:
            return end_most_indel.pos - 1
        else:
            return end_most_indel.pos + 1
    else:
        if left:
            return peak_pos - 1
        else:
            return peak_pos + 1


def get_end_most_indel(indexed_contig, target):
    for k, v in indexed_contig.items():
        if len(v[0]) != len(v[1]):
            return Variant(target.chrom, k, v[0], v[1], target.reference)



#!/usr/bin/env python
# 786
# Aldy source: __init__.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from .version import __version__  # noqa



#cython: profile=False

from .utilities import *
from .localn import make_aligner, align, findall_indels
from pysam.libcfaidx cimport FastaFile
from pysam.libcbcf cimport VariantFile


cdef class NullVariant:
    """This class is returned when :class:`~indelpost.VariantAlignment` cannot find the target indel in the BAM file. Boolean expression evaluates to `False <https://docs.python.org/3/library/stdtypes.html#boolean-values>`__. Ref and Alt alleles are the 
    reference base at the locus.
    
    Parameters
    ----------
    chrom : string
        contig name

    pos : integer
        1-based genomic position

    reference : pysam.FastaFile
        reference FASTA file supplied as
        `pysam.FastaFile <https://pysam.readthedocs.io/en/latest/api.html#pysam.FastaFile>`__ object.
    """
    def __init__(self, str chrom, int pos, FastaFile reference):
        self.chrom = chrom
        self.pos = pos
        self.ref = reference.fetch(chrom, pos - 1, pos)
        self.alt = self.ref
        self.reference = reference

    def __getstate__(self):
        return (self.chrom, self.pos, self.ref, self.alt, self.reference.filename)

    def __setstate__(self, state):
        self.chrom = state[0]
        self.pos = state[1]
        self.ref = state[2]
        self.alt = state[3]

        self.reference = FastaFile(state[4])

    def __bool__(self):
        return False
    
    def __eq__(self, other):
        
        if isinstance(other, Variant):
            return False
       
        chrom_equal = (self.chrom == other.chrom)
        pos_equal = (self.pos == other.pos)
        ref_equal = (self.ref == other.ref)
        alt_equal = (self.alt == other.alt)
        return all([chrom_equal, pos_equal, ref_equal, alt_equal])

    def __hash__(self):
        hashable = (self.chrom, self.pos, self.ref, self.alt)
        return hash(hashable)


cdef class Variant:
    """This class accepts a VCF-style variant representation as input. 
    Equality holds between :class:`~indelpost.Variant` objects 
    if they are indentical in the normalized form.  

    Parameters
    ----------
    chrom : string
        contig name. 
    
    pos : integer
        1-based genomic position.
    
    ref : string
        VCF-style reference allele. 
    
    alt : string
        VCF-style alternative allele.
    
    reference : pysam.FastaFile
        reference FASTA file supplied as 
        `pysam.FastaFile <https://pysam.readthedocs.io/en/latest/api.html#pysam.FastaFile>`__ object.
    
    Raises
    ------
    ValueError
        if the input locus is not defined in the reference or the input alleles contain letters 
        other than A/a, C/c, G/g, T/t, and N/n. 
        
    """
    def __init__(self, str chrom, int pos, str ref, str alt, FastaFile reference, skip_validation=False):
        self._chrom = chrom
        self.pos = pos
        self.ref = ref
        self.alt = alt
        self.reference = reference

        
        if not skip_validation:
            self.chrom = self.__format_chrom_name(self._chrom, reference=reference)
            self.__validate()
        else:
            self.chrom = chrom
        
    def __getstate__(self):
        return (self.chrom, self.pos, self.ref, self.alt, self.reference.filename) 
        

    def __setstate__(self, state):        
        self.chrom = state[0]
        self.pos = state[1]
        self.ref = state[2]
        self.alt = state[3]
        
        self.reference = FastaFile(state[4])
    
    
    def __format_chrom_name(self, chrom, **kwargs):
        if kwargs.get("vcf", False):
            chrom_names = list(kwargs["vcf"].header.contigs)
        elif kwargs.get("reference", False):
            chrom_names = kwargs["reference"].references

        is_prefixed = True if chrom_names[0].startswith("chr") else False
        is_mt = True if "chrMT" in chrom_names or "MT" in chrom_names else False

        chrom = chrom.replace("chr", "")
        if chrom == "M" and is_mt:
            chrom = "MT"
        elif chrom == "MT" and not is_mt:
            chrom = "M"

        if is_prefixed:
            chrom = "chr" + chrom

        return chrom


    cpdef __validate(self):
        if not self.ref or not self.alt:
            raise ValueError("Allele may not be empty")

        if self.ref == self.alt:
            raise ValueError(
                "Not a variant: reference allele and alternate allele may not be identical"
            )

        bases = {"A", "C", "T", "G", "N", "a", "t", "c", "g", "n"}
        ref_lst, alt_lst = list(self.ref), list(self.alt)
        if not set(ref_lst) <= bases or not set(alt_lst) <= bases:
            self.ref = "".join([base if base in bases else "N" for base in ref_lst])
            self.alt = "".join([base if base in bases else "N" for base in alt_lst])

        # check if contig is valid
        try:
            if not self.reference.fetch(self.chrom, self.pos - 1 , self.pos):
                raise ValueError("The locus is not defined in the reference")
        except:
            raise ValueError("The locus is not defined in the reference")
    
    @property
    def variant_type(self):
        """ returns "I" if the net allele-length change is gain, "D" if loss, 
            "S" if signle-nucleotide substitution (zero net change), "M" if multi-nucleotide substitution 
            (zero net change). 
        """ 
        
        cdef int r_len, a_len
        cdef str var_type
            
        r_len, a_len = len(self.ref), len(self.alt)
        if r_len < a_len:
            var_type = "I"
        elif r_len > a_len:
            var_type = "D"
        elif a_len == 1:
            var_type = "S"
        else:
            var_type = "M"

        return var_type


    @property
    def is_del(self):
        """evaluates if :attr:`~indelpost.Variant.variant_type` is "D".
        """
        return self.variant_type == "D"


    @property
    def is_ins(self):
        """evaluates if :attr:`~indelpost.Variant.variant_type` is "I".
        """
        return self.variant_type == "I"

        
    @property
    def is_indel(self):
        """returns True if :attr:`~indelpost.Variant.variant_type` is "I" or "D".
        """
        return self.is_ins or self.is_del


    @property
    def indel_seq(self):
        """returns the inserted/deleted sequence for non-complex indels. None for substitutions.
        """
        if self.is_ins:
            return self.alt[len(self.ref) :]
        elif self.is_del:
            return self.ref[len(self.alt) :]
        else:
            return ""


    def __eq__(self, other):
        
        if isinstance(other, NullVariant):
            return False
        
        cdef Variant i, j
        i, j = self.normalize(), other.normalize()

        chrom_eq = (
            (i.chrom.replace("chr", "") == j.chrom.replace("chr", ""))
            or
            (i._chrom.replace("chr", "") == j._chrom.replace("chr", ""))
        )
             
        equivalent = (
                chrom_eq
                and i.pos == j.pos
                and j.ref.upper() == i.ref.upper()
                and i.alt.upper() == j.alt.upper()
        )
        
        return equivalent


    def __hash__(self):
        cdef Variant i = self.normalize() if self.is_indel else self
        hashable = (i._chrom, i.pos, i.ref, i.alt)

        return hash(hashable)


    def __dealloc__(self):
        pass


    @property
    def is_leftaligned(self):
        """returns True if the indel position is the smallest possible value (left-aligned). 
        """
        if self.ref[-1].upper() != self.alt[-1].upper():
            return True
        elif "N" in self.ref.upper() or "N" in self.alt.upper():
            return True

    
    @property
    def is_normalized(self):
        """returns True if left-aligned and the allele representations are minimal. 
        """
        if self.is_leftaligned:
            if len(self.ref) > 1 and len(self.alt) and (self.ref[0].upper() == self.alt[0].upper()):
                return False
            else:
                return True
        else:
            return False


    def normalize(self, inplace=False):
        """normalizes :class:`~indelpost.Variant` object.
        
        Parameters
        ----------

        inplace : bool
            returns None and normalizes this :class:`~indelpost.Variant` object if True. 
            Otherwise, returns a normalized copy of this object. Default to False.

        """
        cdef Variant i
        cdef str ihs
        cdef int n = 0, lhs_len = 0


        if inplace:
            i = self
        else:
            i = Variant(self.chrom, self.pos, self.ref, self.alt, self.reference, skip_validation=True)
        
        condition_1 = i.ref[-1].upper() == i.alt[-1].upper() != "N" 
        lhs = i.reference.fetch(i.chrom, max(0, i.pos - 1 - 300), i.pos - 1)[::-1]
        lhs_len = len(lhs)
        while condition_1 and n < lhs_len:
            
            left_base = lhs[n]

            i.ref = left_base + i.ref[:-1]
            i.alt = left_base + i.alt[:-1]
            i.pos -= 1
            
            condition_1 = i.ref[-1].upper() == i.alt[-1].upper() != "N"

            n += 1 

        condition_2 = i.ref[0].upper() == i.alt[0].upper()
        condition_3 = len(i.ref) > 1 and len(i.alt) > 1
        while condition_2 and condition_3:
            i.ref = i.ref[1:]
            i.alt = i.alt[1:]
            i.pos += 1
            condition_2 = i.ref[0].upper() == i.alt[0].upper()
            condition_3 = len(i.ref) > 1 and len(i.alt) > 1

        if inplace:
            return None
        else:
            return i
    
    
    def generate_equivalents(self):
        """generates non left-aligned copies of :class:`~indelpost.Variant` object.
        """ 
        
        i = Variant(self.chrom, self.pos, self.ref, self.alt, self.reference, skip_validation=True).normalize()
        
        pos, ref, alt = i.pos, i.ref, i.alt
        is_ins = i.is_ins
        
        res = [i]
        if not i.is_indel:
            return res
        
        
        window = 300
        ref_lim = i.reference.get_reference_length(i.chrom)
        if i.is_non_complex_indel() and i.variant_type == "I":
            rt_flank = i.reference.fetch(i.chrom, i.pos, min(i.pos + window, ref_lim))
        else:
            if i.is_non_complex_indel() and i.variant_type == "D":
                event_len = len(i.indel_seq)
            else:
                event_len = len(i.ref) - 1  
            rt_flank = i.reference.fetch(i.chrom, i.pos + event_len, min(i.pos + event_len + window, ref_lim))

        n = 0
        while self == i and n < window:
            right_base = rt_flank[n]
            if is_ins:
                ref = alt[1]
                alt = alt[1:] + right_base
            else:
                alt = ref[1]
                ref = ref[1:] + right_base
            
            pos += 1 
            
            i = Variant(self.chrom, pos, ref, alt, self.reference, skip_validation=True)
            
            if self == i: 
                res.append(i)
            
            n += 1

        return res


    def _generate_equivalents_private(self):
        
        if self.is_non_complex_indel():
            return self.generate_equivalents()
        else:
            # define complex indel at the start and end of the deleted sequence 
            i = Variant(self.chrom, self.pos, self.ref, self.alt, self.reference, skip_validation=True)
            j = Variant(self.chrom, self.pos + len(self.ref), self.ref, self.alt, self.reference, skip_validation=True)
            return [i, j]
    

    def _get_indel_seq(self, how=None):
        if self.is_non_complex_indel():
            return self.indel_seq
        else:
            if how == "I":
                return self.alt[1:]
            elif how == "D":
                return self.ref[1:]
    
    def _reduce_complex_indel(self, to=None):
        if self.is_non_complex_indel():
            return NullVariant(self.chrom, self.pos, self.reference)
        else:
            if to == "I":
                return Variant(self.chrom, self.pos, self.alt[0], self.alt, self.reference, skip_validation=True)
            elif to == "D":
                return Variant(self.chrom, self.pos, self.ref, self.ref[0], self.reference, skip_validation=True)
         
    
    def query_vcf(self, VariantFile vcf, matchby="normalization", window=50, indel_only=True, as_dict=True):
        """returns a `list <https://docs.python.org/3/library/stdtypes.html#list>`__ of VCF records matching 
        this :class:`~indelpost.Variant` object as `dictionary <https://docs.python.org/3/library/stdtypes.html#dict>`__ (default).

        Parameters
        ----------
        vcf : pysam.VariantFile
            VCF file to be queried. 
            Supply as 
            `pysam.VariantFile <https://pysam.readthedocs.io/en/latest/api.html#pysam.VariantFile>`__ object.

        matchby : string
            - "normalization" (default) matches by normalizing VCF records.
            - "locus" matches by the normalized genomic locus.  
            - "exact" finds the exact matche without normalization. 
                
        window : integer
            searches the VCF records in indel position +/- window.
            
        indel_only : bool
            returns matched indel and SNV records if False. Meaningful when matchby is "locus".
               
        as_dict : bool
            returns `pysam.VariantRecord <https://pysam.readthedocs.io/en/latest/api.html#pysam.VariantRecord>`__ if False.
        """
        matchbys = ["normalization", "locus", "exact"]
        if not matchby in matchbys:
            raise ValueError("match by one of: %s" % matchbys)
        
        if self.variant_type == "S":
            leftaligned_pos, window = self.pos, 1
        else:
            leftaligned_pos = self.normalize().pos
        
        chrom = self.__format_chrom_name(self.chrom, vcf=vcf)
        
        searchable = vcf.fetch(chrom, leftaligned_pos - 1, leftaligned_pos - 1 + window)
        
        if not searchable:
            return []

        records = to_flat_list(
            [
                to_flat_vcf_records(rec)
                for rec in searchable
            ]
        )
        
        hits = [
                record.orig
                for record in records
                if match_indels(
                    Variant(self.chrom, record.pos, record.ref, record.alt, self.reference), 
                    self, 
                    matchby,
                    indel_only,
                )
        ]
        
        if as_dict:
            hits = [
                {
                    "CHROM": hit.chrom,
                    "POS": hit.pos,
                    "ID": hit.id,
                    "REF": hit.ref,
                    "ALT": ",".join(list(hit.alts)),
                    "QUAL": hit.qual,
                    "FILTER": to_dict(hit.filter),
                    "INFO": to_dict(hit.info),
                    "FORMAT": to_dict(hit.format),
                    "SAMPLES": to_dict(hit.samples),
                }
                for hit in hits
            ]
        
        return hits


    def left_flank(self, window=50, normalize=False):
        """extracts the left-flanking reference sequence. See also :meth:`~indelpost.Variant.right_flank`.

        Parameters
        ----------
        window : integer
            extract the reference sequence [variant_pos - window, variant_pos]. 
        normalize : bool
            if True, the normalized indel position is used as the end of the flanking sequence.
        """ 
        if normalize:
            i = Variant(self.chrom, self.pos, self.ref, self.alt, self.reference, skip_validation=True)
        else:
            i = self

        if i.is_non_complex_indel():
            pos = i.pos
        else:
            pos = i.pos - 1
        
        lt_flank = i.reference.fetch(i.chrom, max(0, pos - window), pos)
        
        return lt_flank

    
    def right_flank(self, window=50, normalize=False):
        """extracts the right-flanking reference sequence. See also :meth:`~indelpost.Variant.left_flank`. 

        Parameters
        ----------
        window : integer
            extract the reference sequence [variant_end_pos, variant_end_pos + window].
        normalize : bool
            if True, the normalized indel position is used as the start of the flanking sequence.
        """
        if normalize:
            i = Variant(self.chrom, self.pos, self.ref, self.alt, self.reference, skip_validation=True)
        else:
            i = self

        ref_lim = i.reference.get_reference_length(i.chrom)
        if i.is_non_complex_indel() and i.variant_type == "I":
            rt_flank = i.reference.fetch(i.chrom, i.pos, min(i.pos + window, ref_lim))
        else:
            if i.is_non_complex_indel() and i.variant_type == "D":
                event_len = len(i.indel_seq)
            else:
                event_len = len(i.ref) - 1  
            rt_flank = i.reference.fetch(i.chrom, i.pos + event_len, min(i.pos + event_len + window, ref_lim))

        return rt_flank

    
    def count_repeats(self, by_repeat_unit=True):
        """counts indel repeats in the flanking reference sequences. The search window is
        defined by :meth:`~indelpost.Variant.left_flank` and :meth:`~indelpost.Variant.right_flank`.
        
        Parameters
        ----------
        by_repeat_unit : bool
            count by the smallest tandem repeat unit. For example, the indel sequence "ATATATAT" has
            tandem units "ATAT" and "AT". The occurrence of "AT" will be counted if True (default).
        """ 

        if self.is_non_complex_indel():
            seq = self.indel_seq
        else:
            seq = self.alt
        
        if by_repeat_unit:
            seq = to_minimal_repeat_unit(seq)

        lt_flank = self.left_flank()
        lr_repeat = repeat_counter(seq, lt_flank[::-1])
        rt_flank = self.right_flank()
        rt_repeat = repeat_counter(seq, rt_flank)
        
        return lr_repeat + rt_repeat
    

    def is_non_complex_indel(self):
        """returns True only if non-complex indel (False if complex indel or substitution).
        """
        i = self.normalize()
        ref, alt = i.ref, i.alt
        if len(ref) == len(alt):
            return False
        
        if ref[0] != alt[0]:
            return False
        
        the_shorter = ref if i.is_ins else alt
        if len(the_shorter) > 1:
            return False
        
        return True


    def decompose_complex_variant(self, match_score=3, mismatch_penalty=2, gap_open_penalty=4, gap_extension_penalty=0):
        """returns a `list <https://docs.python.org/3/library/stdtypes.html#list>`__ of 
        non-complex :class:`~indelpost.Variant` objects decomposed by the Smith-Waterman local alignment 
        with a given set of score/penalty.

        Parameters
        ----------
        match_score : integer
            default to 3.
        mismatch_penalty : integer
            default to 2.
        gap_open_penalty : integer
            default to 4.
        gap_extension_penalty : integer
            default to 0.    
        """
        if self.is_non_complex_indel():
            return [self]
        
        var = Variant(self.chrom, self.pos, self.ref, self.alt, self.reference, skip_validation=True).normalize()
        
        lt_pos = var.pos - 1
        rt_pos = var.pos - 1 + len(var.ref)
        
        window = 100
        mut_seq = self.reference.fetch(var.chrom, lt_pos - window, lt_pos) + var.alt + self.reference.fetch(var.chrom, rt_pos, rt_pos + window)
        ref_seq = self.reference.fetch(var.chrom, lt_pos - window, lt_pos + len(var.ref) + window)
        
        aln = align(make_aligner(ref_seq, match_score, mismatch_penalty), mut_seq, gap_open_penalty, gap_extension_penalty) 
        
        genome_aln_pos = lt_pos + 1 - window + aln.reference_start
        
        indels, snvs = findall_indels(aln, genome_aln_pos, ref_seq, mut_seq, report_snvs=True)
        
        variants = []
        if indels:
            for idl in indels:
                padding_base = idl["lt_ref"][-1]
                if idl["indel_type"] == "D":
                    ref = padding_base + idl["del_seq"]
                    alt = padding_base
                else:
                    ref = padding_base
                    alt = padding_base + idl["indel_seq"]

                variants.append(Variant(self.chrom, idl["pos"], ref, alt, self.reference, skip_validation=True))
        
        if snvs:
            for snv in snvs:
                variants.append(Variant(self.chrom, snv["pos"], snv["ref"], snv["alt"], self.reference, skip_validation=True))
             
        return variants



# 786
# Aldy source: minor.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.

from typing import List, Set, Tuple, Dict
from natsort import natsorted
import collections
import os

from . import lpinterface
from .common import log, json, Timing
from .gene import Mutation, Gene
from .coverage import Coverage
from .solutions import MajorSolution, SolvedAllele, MinorSolution
from .diplotype import estimate_diplotype


def estimate_minor(
    gene: Gene,
    coverage: Coverage,
    major_sols: List[MajorSolution],
    solver: str,
    max_solutions: int = 1,
    novel: bool = False,
) -> List[MinorSolution]:
    """
    Estimate the optimal minor star-allele.

    :param gene: Gene instance.
    :param coverage: Read coverage instance.
    :param major_sol: Major allele solution for minor star-allele calling.
    :param solver: ILP solver (see :py:mod:`aldy.lpinterface` for supported solvers).
    :param max_solutions: Maximum number of solutions to report.
        Default: 1.
    :param novel: Include non-database functional and silent mutations in the model.
        Default: `False`.
    """

    # Get the list of potential alleles and mutations
    alleles: List[SolvedAllele] = list()
    # Consider all major and minor mutations
    # *from all available major solutions* together
    mutations: Set[Mutation] = set()
    for major_sol in major_sols:
        for sa in major_sol.solution:
            alleles += [
                SolvedAllele(gene, sa.major, mi) for mi in gene.alleles[sa.major].minors
            ]
            mutations |= set(gene.alleles[sa.major].func_muts)
            for minor in gene.alleles[sa.major].minors.values():
                mutations |= set(minor.neutral_muts)
        mutations |= set(major_sol.added)
    mutations |= gene.random_mutations

    # Filter out low quality mutations
    def default_filter_fn(cov, mut):
        # TODO: is this necessary?
        r = gene.region_at(mut.pos)
        if mut.op != "_" and not (
            mut in mutations
            or (r and r[1][0] == "e")
            or (r and r[1] in ["utr3", "utr5", "up"])
        ):
            return False
        cond = cov.basic_filter(mut, cn=coverage.profile.cn_max)
        if mut.op != "_":
            cond = cond and cov.basic_filter(
                mut, cn=major_sol.cn_solution.position_cn(mut.pos) + 0.5
            )
        return cond

    cov = coverage.filtered(Coverage.quality_filter)
    cov = cov.filtered(default_filter_fn)

    if novel:
        for pos, c in cov._coverage.items():
            for m in c:
                if m != "_" and Mutation(pos, m) not in mutations:
                    r = gene.region_at(pos)
                    log.info(
                        "[minor] novel {} ({}; coverage= {:.0f}; func= {})",
                        r[1] if r else "-",
                        cov.percentage(Mutation(pos, m)),
                        gene.is_functional((pos, m)),
                    )
                    mutations.add(Mutation(pos, m))

    # Group by CN solutions
    minor_sols: List[MinorSolution] = []
    cn_sols = {m.cn_solution for m in major_sols}
    min_score = min(m.score for m in major_sols)
    for c in sorted(cn_sols, key=lambda x: x._solution_nice()):
        log.debug("*" * 80)
        majors = [m for m in major_sols if m.cn_solution == c]
        _print_candidates(gene, alleles, c, cov, mutations)
        for major_sol in natsorted(majors, key=lambda s: str(s.solution)):
            sols = solve_minor_model(
                gene,
                cov,
                major_sol,
                alleles,
                mutations,
                solver,
                max_solutions,
            )
            for s in sols:
                s.score += major_sol.score - min_score
            minor_sols += sols
    return minor_sols


def solve_minor_model(
    gene: Gene,
    coverage: Coverage,
    major_sol: MajorSolution,
    alleles_list: List[SolvedAllele],
    mutations: Set[Mutation],
    solver: str,
    max_solutions: int = 1,
) -> List[MinorSolution]:
    """
    Solves the minor star-allele detection problem via integer linear programming.

    :param gene: Gene instance.
    :param coverage: Read coverage instance.
    :param major_sol: Major allele solution for minor star-allele calling.
    :param alleles_list: Candidate minor star-alleles.
    :param mutations: Mutations to consider for model building (all other mutations
        are ignored).
    :param solver: ILP solver (see :py:mod:`aldy.lpinterface` for supported solvers).
    :param max_solutions: Maximum number of solutions to report.
        Default: 1.

    .. note::
        Please see `Aldy paper <https://www.nature.com/articles/s41467-018-03273-1>`_
        (section Methods/Genotype refining) for the model explanation.
        Currently returns only the first optimal solution.
    """

    log.debug("[minor] major= {}", major_sol._solution_nice())
    model = lpinterface.model("AldyMinor", solver)
    debug_info = json[gene.name]["minor"][len(json[gene.name]["minor"])]

    # Establish minor alleles and their mutations
    alleles: Dict[Tuple[SolvedAllele, int], Set[Mutation]] = {
        (a, 0): set(gene.alleles[a.major].func_muts)
        | set(gene.alleles[a.major].minors[a.minor].neutral_muts)
        for a in alleles_list
    }

    for a, _ in list(alleles):
        max_cn = major_sol.solution[SolvedAllele(gene, a.major, "", a.added, a.missing)]
        for cnt in range(1, max_cn):
            alleles[a, cnt] = alleles[a, 0]

    VA = {
        a: model.addVar(vtype="B", name=f"A_{a[0].major}_{a[0].minor}_{a[1]}")
        for a in alleles
    }
    for a, cnt in alleles:
        if cnt > 0:
            model.addConstr(VA[a, cnt] <= VA[a, cnt - 1], name=f"CORD_{a.minor}_{cnt}")

    # Make sure that the sum of all subaleles matches the count
    # of the corresponding major alleles
    for sa, cnt in major_sol.solution.items():
        expr = model.quicksum(
            v
            for (vs, _), v in VA.items()
            if (vs.major, vs.added, vs.missing) == (sa.major, sa.added, sa.missing)
        )
        model.addConstr(expr <= cnt, name=f"CCNT_{sa.major}_1")
        model.addConstr(expr >= cnt, name=f"CCNT_{sa.major}_2")
    # Disable other alleles
    model.addConstr(
        model.quicksum(VA.values()) <= sum(major_sol.solution.values()),
        name="CCNT_OTHER",
    )

    # Add a binary variable for each allele/mutation pair, where mutation belongs
    # to that allele, that will indicate whether such mutation will be kept or not.
    # Second variable stands for the corresponding VA * VKEEP binary product
    # (binary product transformation).
    VKEEP: dict = {
        a: {
            m: (
                model.addVar(
                    vtype="B", name=f"K_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}"
                ),
                model.addVar(
                    vtype="B",
                    name=f"MUL_K_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}",
                ),
            )
            for m in alleles[a]
        }
        for a in alleles
    }
    # Add a binary variable for each allele/mutation pair, where mutation DOES NOT
    # belong to that allele,that will indicate whether such mutation will be assigned
    # to that allele or not.
    # Second variable stands for the corresponding VA * VNEW product
    # (binary product transformation).
    VNEW: dict = {
        a: {
            m: (
                model.addVar(
                    vtype="B", name=f"N_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}"
                ),
                model.addVar(
                    vtype="B",
                    name=f"MUL_N_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}",
                ),
            )
            for m in mutations
            if gene.has_coverage(a[0].major, m.pos) and m not in alleles[a]
        }
        for a in alleles
    }

    # Add an error variable for each mutation, and populate error constraints
    VERR = {
        m: model.addVar(lb=-model.INF, ub=model.INF, name=f"E_{m.pos}_{m.op}")
        for m in mutations
    }
    constraints = {m: 0 for m in mutations}
    for m in mutations:
        for a in alleles:
            if m in alleles[a]:
                constraints[m] += model.prod(VKEEP[a][m][1], [VA[a], VKEEP[a][m][0]])
            # Add this *only* if CN of this region in a given allele is positive
            elif gene.has_coverage(a[0].major, m.pos):
                constraints[m] += model.prod(VNEW[a][m][1], [VA[a], VNEW[a][m][0]])

    # Fill the constraints for non-variations (i.e. reference genome matches)
    for pos in set(m.pos for m in constraints):
        ref_m = Mutation(pos, "_")  # type: ignore
        VERR[ref_m] = model.addVar(lb=-model.INF, ub=model.INF, name=f"E_{pos}_REF")
        constraints[ref_m] = 0
        for a in alleles:
            if not gene.has_coverage(a[0].major, pos):
                continue
            # Does this allele contain any mutation at the position `pos`?
            # Insertions are not counted as they always contribute to `_`.
            present_muts = [m for m in alleles[a] if m.pos == pos and m[1][:3] != "ins"]
            assert len(present_muts) < 2
            if len(present_muts) == 1:
                constraints[ref_m] += VA[a] - VKEEP[a][present_muts[0]][1]
            else:
                constraints[ref_m] += VA[a]
                muts = [m for m in VNEW[a] if m.pos == pos and m[1][:3] != "ins"]
                for m in muts:
                    constraints[ref_m] -= VNEW[a][m][1]
                # We ensure that only one additional mutation can be selected here
                model.addConstr(
                    model.quicksum(VNEW[a][m][0] for m in muts) <= 1,
                    name=f"CONE_{pos}_{a[0].major}_{a[0].minor}_{a[1]}",
                )

    # Ensure that each constraint matches the observed coverage
    debug_info["cn"] = dict(major_sol.cn_solution.solution)
    debug_info["major"] = {s.major: v for s, v in major_sol.solution.items()}
    debug_info["data"] = []
    for m, expr in sorted(constraints.items()):
        scov = coverage.single_copy(m, major_sol.cn_solution)
        # If scov = 0, no mutations should be selected at that locus
        # (enforced by other constraints)
        cov = coverage[m] / scov if scov > 0 else 0
        model.addConstr(expr + VERR[m] >= cov, name=f"CCOV_{m.pos}_{m.op}")
        model.addConstr(expr + VERR[m] <= cov, name=f"CCOV_{m.pos}_{m.op}")
        debug_info["data"].append((m.pos, m.op, coverage[m]))

    # TODO: downscale ambiguous mutations (right now score == 1 for all mutations)
    score = {}
    for m, v in VERR.items():
        s = 1
        # ops = coverage._coverage[m.pos].get(m.op, [])
        # if ops:
        #     mx = max(mq for mq, _ in ops)
        #     sx = sum(math.log(1 + mq) / math.log(1 + mx) for mq, _ in ops) / len(ops)
        #     s += 1 - sx
        score[model.varName(v)] = s

    # Enforce the following rules:
    # 1) Each mutation is assigned only to alleles that are present in the solution
    for a, mv in VKEEP.items():
        for m, v in mv.items():
            model.addConstr(
                v[0] <= VA[a],
                name=f"CVK_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}",
            )
    for a, mv in VNEW.items():
        for m, v in mv.items():
            model.addConstr(
                v[0] <= VA[a],
                name=f"CVN_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}",
            )
    # 2) Each allele must express ALL of its functional mutations
    for a in alleles:
        for m in alleles[a]:
            if gene.is_functional(m):
                assert m in VKEEP[a]
                model.addConstr(
                    VKEEP[a][m][0] >= VA[a],
                    name=f"CFUNC_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}",
                )
    # 3) No allele can include a mutation with coverage 0
    for a in VKEEP:
        for m, v in VKEEP[a].items():
            if not gene.has_coverage(a[0].major, m.pos):
                model.addConstr(
                    v[0] <= 0,
                    name=f"CZERO_{m.pos}_{m.op}_{a[0].major}_{a[0].minor}_{a[1]}",
                )
    # 4) Avoid extra mutations if there is an existing mutation at the corresponding
    #    locus (either from the definition or added via VNEW)
    for pos in set(m.pos for m in constraints):
        for a in alleles:
            mp = [VKEEP[a][m][1] for m in VKEEP[a] if m.pos == pos]
            ma = [VNEW[a][m][1] for m in VNEW[a] if m.pos == pos]
            # TODO: add support for extra insertions!
            if len(ma) > 1:  # useless if == 1 since variables are binary
                model.addConstr(
                    model.quicksum(ma) <= 1,
                    name=f"CSINGLE_{pos}_{a[0].major}_{a[0].minor}_{a[1]}",
                )
            if len(ma) + len(mp) > 1:
                model.addConstr(
                    model.quicksum(mp + ma) <= 1,
                    name=f"CSINGLEFULL_{pos}_{a[0].major}_{a[0].minor}_{a[1]}",
                )
    # 5) Make sure that each mutation copy has at least one read supporting it
    for m in mutations:
        expr = model.quicksum(VKEEP[a][m][1] for a in alleles if m in VKEEP[a])
        expr += model.quicksum(VNEW[a][m][1] for a in alleles if m in VNEW[a])
        if major_sol.cn_solution.position_cn(m.pos) == 0 or coverage[m] == 0:
            model.addConstr(expr <= 0, name=f"CNOCOV_{m.pos}_{m.op}")
        else:
            model.addConstr(expr <= coverage[m], name=f"CMAXCOV_{m.pos}_{m.op}")
            # Ensure that at least one allele picks an existing non-filtered mutation
            model.addConstr(expr >= 1, name=f"CMINONE_{m.pos}_{m.op}")
    # 6) Do the same for non-mutations
    for pos in set(m.pos for m in constraints):
        expr = 0
        max_mut = 0
        for a in alleles:
            e = [VKEEP[a][m][1] for m in VKEEP[a] if m.pos == pos]
            e += [VNEW[a][m][1] for m in VNEW[a] if m.pos == pos]
            max_mut = max(max_mut, len(e))
            expr += len(e) * VA[a] - model.quicksum(e)
        if isinstance(expr, int) and expr == 0:
            continue
        m = Mutation(pos, "_")
        if major_sol.cn_solution.position_cn(m.pos) == 0:
            model.addConstr(expr <= 0, name=f"CNOCOV_{m.pos}_{m.op}")
        else:
            # If there is no coverage at `pos` at all despite CN being > 0 (or if there
            # is coverage only on a functional mutation that cannot be selected),
            # allow the allele to select a non-existent non-mutation to prevent
            # infeasibility. Should not happen with "sane" datasets...
            model.addConstr(
                expr
                <= max(major_sol.cn_solution.position_cn(m.pos), coverage[m], max_mut),
                name=f"CMAXCOV_{m.pos}_{m.op}",
            )

    # 7) Respect phasing
    VPHASEERR = []
    VPHASE = {}
    modes = collections.defaultdict(int)
    if coverage.profile.phase and coverage.sam:
        mut_pos = {m.pos for m in mutations}
        for rr, rv in coverage.sam.phases.items():
            c = []
            for k, v in rv.items():
                if k in mut_pos:
                    c.append((k, v))
            c = sorted(c)
            if len(c) > 1:
                modes[tuple(c)] += 1
        log.debug("[minor] number of phases= {}", len(modes))
        log.debug("[minor] number of alleles= {}", len(alleles))
        if len(modes) * len(alleles) > coverage.profile.minor_phase_vars:
            max_sample = len(modes) * (
                coverage.profile.minor_phase_vars / (len(modes) * len(alleles))
            )
            log.debug("[minor] downsampling to= {}", max_sample)
            skip = len(modes) / max_sample
            if max_sample < len(modes):
                mi = list(modes.items())
                modes = dict(mi[i] for i in range(0, len(mi), int(skip)))
        with Timing("[minor] Model setup"):
            addVar = lambda n: model.addVar(vtype="B", name=n, update=False)  # noqa
            vars = 0
            for ri, (rr, cnt) in enumerate(modes.items()):
                r = dict(rr)
                for ai, a in enumerate(alleles):
                    pos, neg, e = [], [], 0
                    for m in mutations:
                        if m.pos not in r or not gene.has_coverage(a[0].major, m.pos):
                            continue
                        if m in VKEEP[a]:
                            (pos if m.op == r[m.pos] else neg).append(VKEEP[a][m][0])
                        elif m in VNEW[a]:
                            (pos if m.op == r[m.pos] else neg).append(VNEW[a][m][0])
                    if len(pos) + len(neg) > 1:
                        v = VPHASE[ai, ri] = addVar(f"PH_{ai}_{ri}")
                        vars += 1
                        model.addConstr(VPHASE[ai, ri] <= VA[a], name=f"PH_{ai}_{ri}")
                        e = 0
                        for i, vm in enumerate(pos):
                            v_vp = model.prod(addVar(f"PHASE2_{ai}_{ri}_{i}"), [v, vm])
                            vars += 1
                            e += v - v_vp
                        for i, vm in enumerate(neg):
                            v_vp = model.prod(addVar(f"PHASE3_{ai}_{ri}_{i}"), [v, vm])
                            vars += 1
                            e += v_vp
                        VPHASEERR.append(cnt * e)
            for ri, _ in enumerate(modes):
                v = [
                    VPHASE[ai, ri] for ai, _ in enumerate(alleles) if (ai, ri) in VPHASE
                ]
                if v:
                    e = model.quicksum(v)
                    model.addConstr(e <= 1, name=f"PHASE4_{ri}_1")
                    model.addConstr(e >= 1, name=f"PHASE4_{ri}_2")
            model.update()
            log.debug("[minor] phasing setup done, vars= {}", vars)

    # Objective: minimize the absolute sum of errors ...
    objective = 0
    o_error = model.abssum((v for v in VERR.values()), coeffs=score)
    objective += o_error
    # ... and penalize the misses ...
    o_penal = coverage.profile.minor_miss * model.quicksum(
        len(VKEEP[a]) * VA[a] for a in VKEEP
    )
    o_penal -= coverage.profile.minor_miss * model.quicksum(
        v[1] for a in VKEEP for _, v in VKEEP[a].items()
    )
    # ... and additions ...
    cnt = 0
    for a in VNEW:
        for _, v in VNEW[a].items():
            # HACK:
            # Add cnt/10000 to select the smallest allele if there is a tie
            o_penal += coverage.profile.minor_add * (1 + cnt / 1000000) * v[0]
            cnt += 1
    # ... and novel functional mutations from the major model...
    for m in {m for a in VNEW for m in VNEW[a]}:
        vars = [
            VNEW[a][m][0]
            for a in VNEW
            if m in VNEW[a]
            if gene.is_functional(m)
            if m not in gene.alleles[a[0].major].func_muts
        ]
        if vars:
            vo = model.addVar(vtype="B", name=f"VNEWOR_{m.pos}_{m.op}")
            model.addConstr(vo <= model.quicksum(vars), name=f"VNEWOR1_{m.pos}_{m.op}")
            for vi, v in enumerate(vars):
                model.addConstr(vo >= v, name=f"VNEWOR2_{m.pos}_{m.op}_{vi}")
            o_penal += coverage.profile.minor_add / 2 * vo
    objective += o_penal
    # ... and phasing error!
    o_phase = coverage.profile.minor_phase * model.quicksum(VPHASEERR)
    objective += o_phase

    model.setObjective(objective)

    # Solve the model
    results = {}
    with Timing("[minor] Model solution"):
        for status, opt, sol in model.solutions():
            solution = []
            for allele, value in VA.items():
                if model.getValue(value) <= 0:
                    continue

                if coverage.profile.phase and coverage.sam:
                    ai = next(i for i, a in enumerate(alleles) if allele == a)
                    _print_phase(
                        model, gene, mutations, modes, VPHASE, VKEEP, VNEW, allele, ai
                    )

                added: List[Mutation] = []
                missing: List[Mutation] = []
                for m, mv in VKEEP[allele].items():
                    if not model.getValue(mv[0]):
                        missing.append(m)
                for m, mv in VNEW[allele].items():
                    if model.getValue(mv[0]):
                        added.append(m)
                    else:
                        copies = coverage.single_copy(m, major_sol.cn_solution)
                        copies = coverage[m] / copies if copies > 0 else 0
                        if abs(copies - major_sol.cn_solution.max_cn()) > 1e-5:
                            continue
                        # HACK: add homozygous mutation to _all_ alleles
                        # Works only if a mutation is unambiguiusly homozygous;
                        # otherwise, it will fall back to the model defaults
                        if m not in alleles[allele]:
                            added.append(m)
                solution.append(
                    SolvedAllele(
                        gene,
                        allele[0].major,
                        allele[0].minor,
                        allele[0].added + added,
                        missing,
                    )
                )

            sol = MinorSolution(
                score=opt,
                solution=solution,
                major_solution=major_sol,
                profile=coverage.profile,
            )
            _ = estimate_diplotype(gene, sol)
            debug_info["sol"] = [(s.minor, s.added, s.missing) for s in solution]
            debug_info["diplotype"] = sol.diplotype
            log.debug(
                f"[minor] status= {status}; opt= {opt:.2f} "
                + "(error= {:.2f}, penal={:.2f}, phase= {:.2f}) "
                + f"solution= {sol._solution_nice()}",
                model.getValue(o_error),
                model.getValue(o_penal),
                model.getValue(o_phase),
            )
            if str(sol) not in results:
                results[str(sol)] = sol
            if len(results) >= max_solutions:
                break
    if not results:
        log.debug("[minor] solution= []")
        debug_info["sol"] = []
        if "ALDY_IIS" in os.environ:  # Set to debug infeasible models
            model.model.computeIIS()
            model.dump("minor.iis.ilp")
    return sorted(results.values(), key=lambda x: str(x.get_minor_diplotype()))


def _print_candidates(gene, alleles, cn_sol, coverage, muts):
    """Pretty-print the list of allele candidates and their mutations."""

    def print_mut(m):
        copies = coverage.single_copy(m, cn_sol)
        copies = coverage[m] / copies if copies > 0 else 0
        impact = gene.get_functional(m)
        return (
            f"  {gene.get_rsid(m):12} {str(m):15} "
            + f"{gene.get_refseq(m, from_atg=True):10} "
            + f"(cov={coverage[m]:4}, cn= {copies:3.1f}; "
            + (f"impact={impact})" if impact else "")
        )

    log.debug("[minor] candidate mutations=")
    for m in sorted(muts):
        if coverage[m]:
            log.debug(print_mut(m))

    log.trace("[minor] candidate alleles=")
    muts = muts.copy()
    for sa in natsorted(alleles, key=lambda x: (x.major, x.minor)):
        am = (
            set(gene.alleles[sa.major].func_muts)
            | set(gene.alleles[sa.major].minors[sa.minor].neutral_muts)
            | set(sa.added)
        )
        miss = sum(1 for a in am if coverage[a] == 0)
        log.trace(f"  *{sa.minor} (major= *{sa.major}; miss= {miss})")
        for m in sorted(am):
            if m in muts:
                muts.remove(m)
            log.trace("  {}", print_mut(m))
    if len(muts) > 0:
        log.trace("  Other mutations:")
        for m in sorted(muts):
            log.trace("  {}", print_mut(m))


def _print_phase(model, gene, mutations, modes, VPHASE, VKEEP, VNEW, allele, ai):
    """Pretty-print phasing results."""
    rds = [
        r
        for ri, r in enumerate(modes.items())
        if (ai, ri) in VPHASE
        if model.getValue(VPHASE[ai, ri])
    ]
    log.trace("[phase] {} (reads= {})", allele[0].minor, len(rds))
    for m in sorted(mutations):
        sm = f"  {m} "
        if m in gene.alleles[allele[0].major].func_muts:
            sm += "*"
        elif m in gene.alleles[allele[0].major].minors[allele[0].minor].neutral_muts:
            sm += "#"
        else:
            sm += " "

        if m in VKEEP[allele]:
            sm += "K" if model.getValue(VKEEP[allele][m][0]) else "-"
        elif m in VNEW[allele]:
            sm += "N" if model.getValue(VNEW[allele][m][0]) else "-"
        else:
            sm += " "

        ph = collections.defaultdict(int)
        for r in rds:
            dr = dict(r[0])
            if m.pos in dr:
                c = dr[m.pos]
                if ">" in c:
                    c = c[2]
                elif c.startswith("del"):
                    c = f"-{len(c)-3}"
                ph[c] += r[1]
        if len(ph):
            p = sorted(ph.items())
            log.trace(
                "[phase]    {}\t{}",
                sm,
                "; ".join(f"{o} ({n})" for o, n in p),
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from setuptools.extension import Extension
from Cython.Build import cythonize
from Cython.Distutils import build_ext
from pysam import get_include as pysam_get_include

extra_compile_args = ["-Wno-unused-function"]
extensions = [
    Extension(
        "aldy.indelpost.variant",
        ["aldy/indelpost/variant.pyx"],
        include_dirs=pysam_get_include(),
        extra_compile_args=extra_compile_args,
    ),
    Extension(
        "aldy.indelpost.utilities",
        ["aldy/indelpost/utilities.pyx"],
        include_dirs=pysam_get_include(),
        extra_compile_args=extra_compile_args,
    ),
    Extension(
        "aldy.indelpost.pileup",
        ["aldy/indelpost/pileup.pyx"],
        include_dirs=pysam_get_include(),
        extra_compile_args=extra_compile_args,
    ),
    Extension(
        "aldy.indelpost.varaln",
        ["aldy/indelpost/varaln.pyx"],
        include_dirs=pysam_get_include(),
        extra_compile_args=extra_compile_args,
    ),
    Extension(
        "aldy.indelpost.contig",
        ["aldy/indelpost/contig.pyx"],
        include_dirs=pysam_get_include(),
        extra_compile_args=extra_compile_args,
    ),
    Extension(
        "aldy.indelpost.local_reference",
        ["aldy/indelpost/local_reference.pyx"],
        include_dirs=pysam_get_include(),
        extra_compile_args=extra_compile_args,
    ),
    Extension(
        "aldy.indelpost.softclip",
        ["aldy/indelpost/softclip.pyx"],
        include_dirs=pysam_get_include(),
        extra_compile_args=extra_compile_args,
    ),
    Extension(
        "aldy.indelpost.localn",
        ["aldy/indelpost/localn.pyx"],
        include_dirs=pysam_get_include(),
        extra_compile_args=extra_compile_args,
    ),
    Extension(
        "aldy.indelpost.gappedaln",
        ["aldy/indelpost/gappedaln.pyx"],
        include_dirs=pysam_get_include(),
        extra_compile_args=extra_compile_args,
    ),
    Extension(
        "aldy.indelpost.sswpy",
        sources=["aldy/indelpost/sswpy.pyx", "aldy/indelpost/ssw.c"],
        extra_compile_args=extra_compile_args,
    ),
]

exec(open("aldy/version.py").read())

setup(
    name="aldy",
    version=__version__,
    description="A tool for allelic decomposition (haplotype reconstruction) "
    + "and exact genotyping of highly polymorphic and structurally variant genes",
    url="http://aldy.csail.mit.edu/",
    author="Ibrahim Numanagić",
    author_email="inumanag@mit.edu",
    download_url="https://github.com/inumanag/aldy/tarball/master",
    license="Aldy/IURTC License. Free for academic/non-commercial use.",
    keywords=["cyp2d6", "adme", "genotyping", "illumina", "pgrnseq", "getrm", "allele"],
    install_requires=[
        "pyyaml",
        "logbook",
        "pysam",
        "pytest",
        "ortools",
        "natsort",
        "mappy",
        "numpy",
        "cython",
    ],
    entry_points={"console_scripts": ["aldy = aldy.__main__:console"]},
    packages=find_packages(),
    package_data={
        "aldy.resources": ["*.rst", "aldy/resources/*.rst"],
        "aldy.resources.genes": [
            "*.yml",
            "aldy/resources/genes/*.yml",
            "aldy/tests/resources/pharmacoscan/*.yml",
        ],
        "aldy.resources.profiles": ["*.yml", "aldy/resources/profiles/*.yml"],
        "aldy.tests.resources": [
            "*.json",
            "aldy/tests/resources/*.json",
            "*.bai",
            "aldy/tests/resources/*.bai",
            "*.bam",
            "aldy/tests/resources/*.bam",
            "*.expected",
            "aldy/tests/resources/*.expected",
            "*.yml",
            "aldy/tests/resources/*.yml",
            "*.profile",
            "aldy/tests/resources/*.profile",
            "*.hg38",
            "aldy/tests/resources/*.hg38",
            "*.vcf",
            "aldy/tests/resources/*.vcf",
            "*.gz",
            "aldy/tests/resources/*.gz",
            "*.tbi",
            "aldy/tests/resources/*.tbi",
        ],
        "aldy.indelpost": [
            "*.pyx",
            "aldy/indelpost/*.pyx",
            "*.pxd",
            "aldy/indelpost/*.pxd",
            "*.h",
            "aldy/indelpost/ssw.h",
            "aldy/indelpost/sse2neon.h",
        ],
    },
    cmdclass={"build_ext": build_ext},
    ext_modules=cythonize(extensions, compiler_directives={"language_level": "3"}),
    test_suite="pytest-runner",
    tests_require=["pytest"],
)



# 786
# Aldy source: test_cn_real.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.

# flake8: noqa

import pytest  # noqa


from aldy.__main__ import main
from aldy.common import script_path, log
from aldy.version import __version__
from .test_full import HEADER, escape_ansi


def assert_show(monkeypatch, expected, query=None, gene="aldy.tests.resources/toy.yml"):
    lines = []

    def log_info(*args):
        s = str.format(*args)
        lines.append(s)  # type: ignore

    monkeypatch.setattr(log, "info", log_info)

    args = ["q", script_path(gene)]
    if query:
        args[1] += "*" + query
    main(args)
    expected = expected.strip()
    lines = escape_ansi("\n".join(lines)).strip()
    assert lines == expected


EXPECTED_SHOW = f"""
{HEADER}
--------------------------------------------------------------------------------
Gene TOY
--------------------------------------------------------------------------------
hg19 genome locus: 20:100000000-100000199 (NG_TEST)
5'
Pseudogenes: TOYP (ID 1)
Aminoacid:
  VRTCTYVYVR
Regions of interest:
     Region:                        TOY                      TOYP
         tmp:    20:100000100-100000110    20:100000000-100000010
          e1:    20:100000110-100000120    20:100000010-100000020
          i1:    20:100000120-100000130    20:100000020-100000030
          e2:    20:100000130-100000140    20:100000030-100000040
          i2:    20:100000140-100000150    20:100000040-100000050
          e3:    20:100000150-100000160    20:100000050-100000060
        down:    20:100000160-100000200    20:100000060-100000100
Structural alleles (deletions, conservations and fusions):
       *1: Standard copy-number configuration
       *4: TOYP fusion until i2
       *5: TOYP conservation after e2
       *6: TOY deletion
Major star-alleles:
  *1:
    Key mutations: none
    Minor alleles: *1.001 (*1),
                   *1.002 (*1B)
  *1C:
    Key mutations: 20:100000105.T>A (-5T>A, NG_TEST:105T>A, functional)
    Minor alleles: *1.003 (*1C)
  *2:
    Key mutations: 20:100000111.delAC (2delAC, NG_TEST:111delAC, frameshift),
                   20:100000119.insTT (10insTT, NG_TEST:119insTT, frameshift)
    Minor alleles: *2.001 (*2)
  *3:
    Key mutations: 20:100000151.C>T (42C>T, NG_TEST:151C>T, functional)
    Minor alleles: *3.001 (*3)
  *5:
    Key mutations: 20:100000111.delAC (2delAC, NG_TEST:111delAC, frameshift)
    Minor alleles: *5.001 (*5)
  *6:
    Key mutations: none
    Minor alleles: *6.001 (*6DEL)
"""


def test_show(monkeypatch):
    assert_show(monkeypatch, EXPECTED_SHOW)


EXPECTED_SHOW_MAJOR = f"""
{HEADER}
Gene TOY, major star-allele TOY*2:
  Structure: 1
  Key mutations: 20:100000111.delAC (2delAC, NG_TEST:111delAC, frameshift),
                 20:100000119.insTT (10insTT, NG_TEST:119insTT, frameshift)
  Minor star-alleles:
    *2.001:
      Legacy name: *2
      Silent mutations: none
"""

EXPECTED_SHOW_MAJOR_2 = f"""
{HEADER}
Gene TOY, major star-allele TOY*1C:
  Structure: 1
  Key mutations: 20:100000105.T>A (-5T>A, NG_TEST:105T>A, functional)
  Minor star-alleles:
    *1.003:
      Legacy name: *1C
      Silent mutations: none
"""


def test_show_major(monkeypatch):
    assert_show(monkeypatch, EXPECTED_SHOW_MAJOR, "2")
    assert_show(monkeypatch, EXPECTED_SHOW_MAJOR_2, "1C")


EXPECTED_SHOW_MINOR = f"""
{HEADER}

Gene TOY, minor star-allele TOY*1.002:
  Structure: 1
  Major star-allele: TOY*1
  Legacy name: *1B
  Key mutations: none
  Silent mutations: 20:100000115.T>A (rs28371732, 6T>A, NG_TEST:115T>A)

"""  # noqa


def test_show_minor(monkeypatch):
    assert_show(monkeypatch, EXPECTED_SHOW_MINOR, "1B")


EXPECTED_SHOW_CN = f"""
{HEADER}
Gene CYP2D6, structural allele CYP2D6*13:
  Structure: CYP2D6   CYP2D7
         up: 0        1
       utr5: 0        1
         e1: 0        1
         i1: 1        0
         e2: 1        0
         i2: 1        0
         e3: 1        0
         i3: 1        0
         e4: 1        0
         i4: 1        0
         e5: 1        0
         i5: 1        0
         e6: 1        0
         i6: 1        0
         e7: 1        0
         i7: 1        0
         e8: 1        0
         i8: 1        0
         e9: 1        0
       utr3: 1        0
        ins: 1        0
        pce: 0        0
        rep: 1        0
"""


def test_show_cn(monkeypatch):
    assert_show(monkeypatch, EXPECTED_SHOW_CN, "13", "aldy.resources.genes/cyp2d6.yml")



#cython: embedsignature=True
#cython: profile=False

import random
import numpy as np
from collections import OrderedDict, namedtuple

from .utilities import *

from .variant cimport Variant
from .local_reference cimport UnsplicedLocalReference

from .consensus import make_consensus


random.seed(123)


cdef class Contig:
    """This class represents a consensus contig assembled from a subset (not all) of reads supporting the target indel.
    """
    def __cinit__(self, Variant target, list pileup, UnsplicedLocalReference unspl_loc_ref, int basequalthresh, int mapqthresh, double low_consensus_thresh=0.7, int donwsample_lim=100):
        self.target = target
        self.pileup = pileup

        self.targetpileup = self.__preprocess(mapqthresh, donwsample_lim)

        if self.targetpileup:
            consensus = make_consensus(self.target, self.targetpileup, basequalthresh)
            if consensus:

                self.splice_pattern = get_local_reference(self.target, consensus[2], 50, unspl_loc_ref, unspliced=False, splice_pattern_only=True)

                self.__make_contig(consensus[0], consensus[1], basequalthresh)

                self.failed = False
            else:
                self.qc_passed = False
                self.failed = True
        else:
            self.qc_passed = False
            self.failed = True


    def __preprocess(self, mapqthresh, donwsample_lim):
        targetpileup = [read for read in self.pileup if read is not None and read["is_target"]]

        self.mapq = 0

        #self.splice_pattern = get_local_reference(self.target, targetpileup, window=50, unspliced=False, splice_pattern_only=True)

        #self.is_target_right_aligned = sum(read.get("target_right_aligned", 0) for read in targetpileup)

        if not targetpileup:
            return targetpileup

        if len(targetpileup) > donwsample_lim:
            targetpileup = random.sample(targetpileup, donwsample_lim)

        self.mapq = np.percentile([read["mapq"] for read in targetpileup], 50)
        self.low_qual_mapping_rate = sum(read["mapq"] < mapqthresh for read in targetpileup) / len(targetpileup)

        return targetpileup


    def __make_contig(self, lt_consensus, rt_consensus, basequalthresh):
        self.__index_by_genome_coord(lt_consensus[0], rt_consensus[0])

        self.lt_reference_seq = ""
        self.lt_target_block_reference_seq = ""
        self.lt_consensus_seq = ""
        self.lt_target_block_consensus_seq = ""
        self.lt_consensus_scores = []
        self.lt_target_block_consensus_scores = []

        self.indel_seq = ""

        self.rt_reference_seq = ""
        self.rt_target_block_reference_seq = ""
        self.rt_consensus_seq = ""
        self.rt_target_block_consensus_seq = ""
        self.rt_consensus_scores = []
        self.rt_target_block_consensus_scores = []

        exon_start, exon_end = -np.inf, np.inf
        if self.splice_pattern:
            for exon in self.splice_pattern:
                if exon[0] <= self.target.pos <= exon[1]:
                    exon_start, exon_end = exon[0], exon[1]

        for k, v in self.contig_dict.items():
            #if k < self.target.pos:
            if k < self.lt_end_pos:
                self.lt_reference_seq += v[0]
                self.lt_consensus_seq += v[1]
                self.lt_consensus_scores.extend([v[2]] * len(v[1]))
                if exon_start <= k:
                    self.lt_target_block_reference_seq += v[0]
                    self.lt_target_block_consensus_seq += v[1]
                    self.lt_target_block_consensus_scores.extend([v[2]] * len(v[1]))

            #elif k == self.target.pos:
            elif k == self.lt_end_pos:
                self.lt_reference_seq += v[0][0]
                self.lt_target_block_reference_seq += v[0][0]

                self.lt_consensus_seq += v[1][0]
                self.lt_target_block_consensus_seq += v[1][0]

                self.lt_consensus_scores.append(v[2])
                self.lt_target_block_consensus_scores.extend([v[2]])

                self.indel_seq = self.target.indel_seq
            #elif k > self.target.pos:
            elif k > self.lt_end_pos:
                self.rt_reference_seq += v[0]
                self.rt_consensus_seq += v[1]
                self.rt_consensus_scores.extend([v[2]] * len(v[1]))
                if k <= exon_end:
                    self.rt_target_block_reference_seq += v[0]
                    self.rt_target_block_consensus_seq += v[1]
                    self.rt_target_block_consensus_scores.extend([v[2]] * len(v[1]))

        self.start = lt_consensus[1]
        self.end = rt_consensus[1]

        self.__profile_non_target_variants()

        self.qc_passed = self.__qc()


    def __index_by_genome_coord(self, lt_index, rt_index):
        self.lt_genomic_index = lt_index
        self.rt_genomic_index = rt_index

        lt_end_pos = next(iter(lt_index))

        #hotfix for right-aln-case
        self.lt_end_pos = lt_end_pos

        # the target may be of low quality ("N")
        if "N" in rt_index[lt_end_pos][1]:
            rt_index[lt_end_pos] = (
                                     rt_index[lt_end_pos][0],
                                     self.target.alt,
                                     rt_index[lt_end_pos][2],
                                     rt_index[lt_end_pos][3]
            )

        genome_indexed_contig = lt_index
        genome_indexed_contig.update(rt_index)
        self.contig_dict = OrderedDict(sorted(genome_indexed_contig.items()))

        target_pos_alleles = self.contig_dict[lt_end_pos]
        ref, alt = target_pos_alleles[0], target_pos_alleles[1]
        if len(ref) < len(alt):
            the_shorter, the_longer = ref, alt
        else:
            the_shorter, the_longer = alt, ref

        self.is_non_complex_at_target_pos = the_longer[: len(the_shorter)] == the_shorter
        self.target_ref = ref[1 :]
        self.target_alt = alt[1 :]


    def __profile_non_target_variants(self):
        non_target_variants = [
            Variant(self.target.chrom, k, v[0], v[1], self.target.reference, skip_validation=True)
            for k, v in self.contig_dict.items()
            if v[0] and v[0] != v[1] and k != self.target.pos
        ]
        self.non_target_indels = [var for var in non_target_variants if var.is_indel]
        self.mismatches = [var for var in non_target_variants if not var.is_indel]

        self.gaps = [
            str(len(var.indel_seq)) + var.variant_type for var in self.non_target_indels
        ]
        self.gaps.append(str(len(self.target.indel_seq)) + self.target.variant_type)


    def __qc(self):

        lt_n, lt_len = self.lt_consensus_seq.count("N"), len(self.lt_consensus_seq)
        rt_n, rt_len = self.rt_consensus_seq.count("N"), len(self.rt_consensus_seq)

        qc_stats ={}

        qc_stats["low_qual_base_frac"] = low_qual_fraction(self.targetpileup)

        qc_stats["clip_rate"] = sum(True for k, v in self.contig_dict.items() if not v[0]) / len(self.contig_dict)

        lt_n_proportion = lt_n / lt_len
        rt_n_proportion = rt_n / rt_len
        qc_stats["n_rate"] = (lt_n + rt_n) / (lt_len + rt_len)

        low_consensus_rate_lt = (
            sum(score < self.low_consensus_thresh for score in self.lt_consensus_scores) / lt_len
        )
        low_consensus_rate_rt = (
            sum(score < self.low_consensus_thresh for score in self.rt_consensus_scores) / rt_len
        )

        qc_stats["low_consensus_rate"] = (low_consensus_rate_lt * lt_len + low_consensus_rate_rt * rt_len) / (lt_len + rt_len)

        self.qc_stats = qc_stats
        if qc_stats["low_qual_base_frac"] > 0.2:
            return False
        #elif qc_stats["clip_rate"] > 0.1: #abolish clip rate filter
        #    return False
        elif qc_stats["n_rate"] > 0.1:
            return False
        elif low_consensus_rate_lt > 0.2 or low_consensus_rate_rt > 0.2:
            return False
        else:
            return True


    def _get_splice_patterns(self):
        spls = self.splice_pattern
        if spls:
            intervals = []
            i, last_idx = 0, len(spls) - 1
            while i < last_idx:
                start = spls[i][1] + 1
                end = spls[i + 1][0] - 1
                intervals.append((start, end))
                i += 1

            return intervals


    def get_alignment(self):
        """shows the contig alignment as `namedtuple <https://docs.python.org/3/library/collections.html#collections.namedtuple>`__.
        ContigAlignment.chrom returns chromosome. ContigAlignment.aln returns an alignment dictionary as
        `OrderedDict <https://docs.python.org/3/library/collections.html#collections.OrderedDict>`__.
        The dictionary key is the position and the value is a `tuple <https://docs.python.org/3/library/stdtypes.html#tuple>`__ as (REF, ALT).
        ContigAlignment.spliced_intervals returns a `list <https://docs.python.org/3/library/stdtypes.html#list>`__
        of spliced intervals involved in the contig. `None <https://docs.python.org/3/c-api/none.html>`__ if the region of interest is not spliced.
        """
        ContigAlignment = namedtuple("ContigAlignment", "chrom aln spliced_intervals")

        data = []
        for k, v in self.contig_dict.items():
            if v[1] and v[0]:
                data.append((k, (v[0], v[1])))

        caln = ContigAlignment(chrom=self.target.chrom, aln=OrderedDict(data), spliced_intervals=self._get_splice_patterns())

        return caln


    def get_phasables(self):
        """returns a `list <https://docs.python.org/3/library/stdtypes.html#list>`__  of :class:`~indelpost.Variant` objects phasable with the target indel.
        """
        phasables = []
        chrom = self.target.chrom
        reference = self.target.reference
        for k, v in self.contig_dict.items():
            if v[1] and v[0] and v[1] != v[0]:
                phasables.append(Variant(chrom, k, v[0], v[1], reference, skip_validation=True))

        return phasables


    def get_reference_seq(self, split=False):
        """returns the reference contig sequence.

        Parameters
        ----------
            split : bool
                splits the reference at the target indel position.
        """
        if self.failed:
            return None

        if split:
            if self.is_non_complex_at_target_pos:
                if self.target.is_del:
                    return self.lt_reference_seq, self.indel_seq, self.rt_reference_seq
                else:
                    return  self.lt_reference_seq, "", self.rt_reference_seq
            else:
                return self.lt_reference_seq, self.target_ref, self.rt_reference_seq
        else:
            if self.target.is_non_complex_indel:
                refseq = (
                    self.lt_reference_seq + self.indel_seq + self.rt_reference_seq
                    if self.target.is_del
                    else self.lt_reference_seq + self.rt_reference_seq
                )
            else:
                return self.lt_reference_seq + self.target_ref + self.rt_reference_seq

            return refseq


    def get_contig_seq(self, split=False):
        """returns the contig sequence.

        Parameters
        ----------
            split : bool
                splits the contig sequence at the target indel position.
        """
        if self.failed:
            return None

        if split:
            if self.is_non_complex_at_target_pos:
                if self.target.is_ins:
                    return self.lt_consensus_seq, self.indel_seq, self.rt_consensus_seq
                else:
                    return self.lt_consensus_seq, "", self.rt_consensus_seq
            else:
                return self.lt_consensus_seq, self.target_alt, self.rt_consensus_seq
        else:
            if self.target.is_non_complex_indel:
                conseq = (
                    self.lt_consensus_seq + self.indel_seq + self.rt_consensus_seq
                    if self.target.is_ins
                    else self.lt_consensus_seq + self.rt_consensus_seq
                )
            else:
                conseq = self.lt_consensus_seq + self.target_alt + self.rt_consensus_seq

            return conseq


cdef class FailedContig:
    """This class is returned when contig assembly has failed. Boolean expression evaluates to `False <https://docs.python.org/3/library/stdtypes.html#boolean-values>`__.

    Parameters
    ----------
    target_not_found : bool
        True when contig was not constructed because the target indel was not found.

    is_low_quality : bool
        True when contig was not constructed because the target indel was only suppported by low quality reads.
        Reads are defined low quality if 15% or more bases have quality score < :attr:`~indelpost.VariantAlignment.base_quality_threshold`.

    failed_anyway : bool
        True when contig was not constructed due to the other reasons.
    """
    def __cinit__(self):
        self.target_not_found = False
        self.is_low_quality = False
        self.failed_anyway = False

    def __bool__(self):
        return False


def compare_contigs(orig_contig, new_contig, target_pos):
    if new_contig.failed:
        return orig_contig

    orig_len = len(orig_contig.get_reference_seq())
    orig_clip_rate = orig_contig.qc_stats["clip_rate"]

    new_len = len(new_contig.get_reference_seq())
    new_clip_rate = new_contig.qc_stats["clip_rate"]

    orig_score = contig_centerness_score(orig_contig, target_pos)
    new_score = contig_centerness_score(new_contig, target_pos)

    if new_clip_rate > 0.1:
        return orig_contig

    condition1 = (new_len <= orig_len)
    condition2 = (new_clip_rate > orig_clip_rate and new_clip_rate >=0.03)
    condition3 = (orig_score < new_score)

    if sum([condition1, condition2, condition3]) >= 2:
        return orig_contig
    else:
        return new_contig


def contig_centerness_score(contig, target_pos):
    lt_cnt, rt_cnt = 0, 0
    for k, v in contig.contig_dict.items():
        if v[0] and v[1]:
            if k <= target_pos:
                lt_cnt += 1
            else:
                rt_cnt += 1

    return 0.5 - min(lt_cnt, rt_cnt) /(lt_cnt + rt_cnt)




# 786
# Aldy source: test_diplotype_synthetic.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from collections import Counter
import pytest  # noqa

from aldy.solutions import SolvedAllele, MinorSolution, MajorSolution, CNSolution
from aldy.diplotype import estimate_diplotype
from aldy.gene import Mutation


def assert_diplotype(gene, test, majors):
    sols = []
    cns = []
    for ma in majors:
        if isinstance(ma, tuple):
            sols.append(SolvedAllele(gene, ma[0], added=[Mutation(*m) for m in ma[1]]))
            cns.append(ma[0])
        else:
            sols.append(SolvedAllele(gene, ma))
    cns = [c for m in cns for c, cc in gene.cn_configs.items() if m in cc.alleles]
    minor = MinorSolution(
        0, sols, MajorSolution(0, Counter(sols), CNSolution(gene, 0, cns), [])
    )
    estimate_diplotype(gene, minor)
    assert test == minor.get_major_diplotype(), "Diplotype not equal"


def test_basic(toy_gene):
    assert_diplotype(toy_gene, "*1 / *1C", ["1", "1C"])


def test_tandem(toy_gene):
    assert_diplotype(toy_gene, "*1C + *4 / *3", ["4", "1C", "3"])


def test_multi(toy_gene):
    assert_diplotype(toy_gene, "*1 + *4 / *3 + *3", ["4", "1", "3", "3"])
    assert_diplotype(
        toy_gene, "*1C + *4 / *3 + *3 + *3 + *3", ["4", "1C", "3", "3", "3", "3"]
    )


def test_combination(toy_gene):
    assert_diplotype(toy_gene, "*1 + *4 / *1C + *4 + *3", ["4", "1C", "1", "3", "4"])


def test_novel(toy_gene):
    assert_diplotype(
        toy_gene,
        "*1 + *4 / *1C+100000151.C>T + *4 + *3",
        ["4", ("1C", [(100_000_150, "C>T")]), "1", "3", "4"],
    )


def test_deletion(toy_gene):
    assert_diplotype(toy_gene, "*2 / *6", ["6", "2"])



# 786
# Aldy source: test_gene.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


import pytest  # noqa

from aldy.gene import Mutation, CNConfig, CNConfigType, MajorAllele, MinorAllele
from aldy.common import GRange


def test_gene_basic(toy_gene):
    assert toy_gene.name == "TOY"
    assert toy_gene.version == "test-2.0 (2020-11-17)"
    assert toy_gene.pharmvar is None
    assert toy_gene.ensembl is None

    assert toy_gene.refseq == "NG_TEST"
    assert toy_gene.seq == "ACGT" * 50, "Sequence"


def test_gene_regions(toy_gene):
    assert toy_gene.genome == "hg19"
    assert toy_gene.chr == "20"
    assert toy_gene.strand == 1
    assert toy_gene.chr_to_ref == {100_000_000 + i: i for i in range(200)}
    assert toy_gene.ref_to_chr == {i: 100_000_000 + i for i in range(200)}

    # Test gene regions
    assert toy_gene.regions == [
        {
            "tmp": GRange("20", 100_000_100, 100_000_110),
            "e1": GRange("20", 100_000_110, 100_000_120),
            "i1": GRange("20", 100_000_120, 100_000_130),
            "e2": GRange("20", 100_000_130, 100_000_140),
            "i2": GRange("20", 100_000_140, 100_000_150),
            "e3": GRange("20", 100_000_150, 100_000_160),
            "down": GRange("20", 100_000_160, 100_000_200),
        },
        {
            "tmp": GRange("20", 100_000_000, 100_000_010),
            "e1": GRange("20", 100_000_010, 100_000_020),
            "i1": GRange("20", 100_000_020, 100_000_030),
            "e2": GRange("20", 100_000_030, 100_000_040),
            "i2": GRange("20", 100_000_040, 100_000_050),
            "e3": GRange("20", 100_000_050, 100_000_060),
            "down": GRange("20", 100_000_060, 100_000_100),
        },
    ], "Regions"
    assert toy_gene.get_wide_region() == ("20", 100_000_000, 100_000_200)
    assert toy_gene.pseudogenes == ["TOYP"], "Pseudogenes"

    coding_seq = "".join(toy_gene.seq[s:e] for [s, e] in toy_gene.exons)
    assert coding_seq == "GTACGTACGTGTACGTACGTGTACGTACGT", "Coding sequence"
    assert toy_gene.aminoacid == "VRTCTYVYVR", "Protein"

    assert toy_gene[100_000_040:100_000_050] == "ACGTACGTAC"
    assert toy_gene[100_000_050] == "G"
    assert toy_gene[100_000_196:100_000_210] == "ACGT" + "N" * 10


def test_gene_alleles(toy_gene):
    def cnify(s):
        s = s.replace(" ", "").replace("|", ",").split(",")
        assert len(s) == len(toy_gene.regions[0]) + len(toy_gene.regions[1])
        i = 0
        cn = [{}, {}]
        for g in range(2):
            for r in toy_gene.regions[g]:
                cn[g][r] = float(s[i])
                i += 1
        return cn

    # Test copy number configurations
    assert toy_gene.do_copy_number
    assert toy_gene.cn_configs == {
        "1": CNConfig(
            cnify("1,  1,1,  1,1,  1,1 | 1,  1,1,  1,1,  1,1"),
            kind=CNConfigType.DEFAULT,
            alleles=set(["1", "1C", "2", "3"]),
            description="Standard copy-number configuration",
        ),
        "4": CNConfig(
            cnify("0,  0,0,  0,1,  1,1 | 1,  1,1,  1,0,  0,0"),
            kind=CNConfigType.LEFT_FUSION,
            alleles=set(["4#1", "4#3"]),
            description="TOYP fusion until i2",
        ),
        "5": CNConfig(
            cnify("1,  1,1,  0,0,  0,0 | 1,  1,1,  2,2,  2,2"),
            kind=CNConfigType.RIGHT_FUSION,
            alleles=set(["5"]),
            description="TOYP conservation after e2",
        ),
        "6": CNConfig(
            cnify("0,  0,0,  0,0,  0,0 | 1,  1,1,  1,1,  1,1"),
            kind=CNConfigType.DELETION,
            alleles=set(["6"]),
            description="TOY deletion",
        ),
    }, "CN configs"
    assert toy_gene.unique_regions == ["e1", "i1", "e2", "i2", "e3"], "Unique regions"

    # Test allele configurations
    assert sorted(toy_gene.alleles.keys()) == [
        "1",
        "1C",
        "2",
        "3",
        "4#1",
        "4#3",
        "5",
        "6",
    ], "Major alleles"
    assert toy_gene.common_tandems == [("1", "4")], "Common tandems"
    assert toy_gene.alleles["1"] == MajorAllele(
        "1",
        cn_config="1",
        minors={
            "1.001": MinorAllele(
                "1.001", "1", activity="normal function", evidence="D"
            ),
            "1.002": MinorAllele(
                "1.002", "1B", neutral_muts=set([Mutation(100_000_114, "T>A")])
            ),
        },
    )
    assert toy_gene.alleles["1C"] == MajorAllele(
        "1C",
        cn_config="1",
        func_muts=set([Mutation(100_000_104, "T>A")]),
        minors={
            "1.003": MinorAllele(
                "1.003", "1C", pharmvar="https://www.pharmvar.org/haplotype/NA"
            )
        },
    )
    assert toy_gene.alleles["2"] == MajorAllele(
        "2",
        cn_config="1",
        func_muts=set([Mutation(100_000_110, "delAC"), Mutation(100_000_118, "insTT")]),
        minors={"2.001": MinorAllele("2.001", "2")},
    )
    assert toy_gene.alleles["3"] == MajorAllele(
        "3",
        cn_config="1",
        func_muts=set([Mutation(100_000_150, "C>T")]),
        minors={
            "3.001": MinorAllele(
                "3.001", "3", neutral_muts=set([Mutation(100_000_147, "insA")])
            )
        },
    )
    assert toy_gene.alleles["4#1"] == MajorAllele(
        "4#1",
        cn_config="4",
        minors={"4#1.001": MinorAllele("4#1.001")},
    )
    assert toy_gene.alleles["4#3"] == MajorAllele(
        "4#3",
        cn_config="4",
        func_muts=set([Mutation(100_000_150, "C>T")]),
        minors={
            "4#3.001": MinorAllele(
                "4#3.001", neutral_muts=set([Mutation(100_000_147, "insA")])
            )
        },
    )
    assert toy_gene.alleles["5"] == MajorAllele(
        "5",
        cn_config="5",
        func_muts=set([Mutation(100_000_110, "delAC")]),
        minors={"5.001": MinorAllele("5.001", "5")},
    )
    assert toy_gene.alleles["6"] == MajorAllele(
        "6",
        cn_config="6",
        minors={"6.001": MinorAllele("6.001", "6DEL")},
    )
    assert toy_gene.deletion_allele() == "6"

    # Test mutations
    assert toy_gene.mutations == {
        (100_000_000 + 105 - 1, "T>A"): ("functional", "-", 105 - 1, 105 - 1, "T>A"),
        (100_000_000 + 111 - 1, "delAC"): (
            "frameshift",
            "-",
            111 - 1,
            111 - 1,
            "delAC",
        ),
        (100_000_000 + 115 - 1, "T>A"): (None, "rs28371732", 115 - 1, 115 - 1, "T>A"),
        (100_000_000 + 119 - 1, "insTT"): (
            "frameshift",
            "-",
            119 - 1,
            119 - 1,
            "insTT",
        ),
        (100_000_000 + 148 - 1, "insA"): (None, "-", 148 - 1, 148 - 1, "insA"),
        (100_000_000 + 151 - 1, "C>T"): ("functional", "-", 151 - 1, 151 - 1, "C>T"),
    }, "Mutations"
    assert toy_gene.has_coverage("1", 100_000_147)
    assert not toy_gene.has_coverage("5", 100_000_147)

    assert toy_gene.get_rsid((100_000_104, "T>A"))


def test_region_at(toy_gene):
    # Test gene auxiliary functions
    assert toy_gene.region_at(100_000_105) == (0, "tmp"), "region_at test"
    assert toy_gene.region_at(100_000_127) == (0, "i1"), "region_at test"
    assert toy_gene.region_at(100_000_155) == (0, "e3"), "region_at test"
    assert toy_gene.region_at(100_000_165) == (0, "down"), "region_at test"
    assert toy_gene.region_at(100_000_265) is None, "region_at test"
    assert toy_gene.region_at(100_000_005) == (1, "tmp"), "region_at test"
    assert toy_gene.region_at(100_000_027) == (1, "i1"), "region_at test"
    assert toy_gene.region_at(100_000_055) == (1, "e3"), "region_at test"
    assert toy_gene.region_at(100_000_080) == (1, "down"), "region_at test"
    assert toy_gene.region_at(90_000_000) is None, "region_at test"


def test_get_functional(toy_gene):
    assert toy_gene.get_functional((100_000_118, "insTT")) == "frameshift"
    assert toy_gene.get_functional((100_000_111, "A>C"), infer=False) is None
    assert toy_gene.get_functional((100_000_111, "A>C"), infer=True) == "V1A"
    assert not toy_gene.is_functional((100_000_111, "A>C"), infer=False)
    assert toy_gene.is_functional((100_000_104, "T>A"))


def test_get_rsid(toy_gene):
    assert toy_gene.get_rsid(100_000_114, "T>A") == "rs28371732"
    assert toy_gene.get_rsid(Mutation(100_000_118, "insTT"), default=False) == "-"
    assert toy_gene.get_rsid(Mutation(100_000_118, "insTT")) == "100000119.insTT"


def test_get_refseq(toy_gene):
    assert toy_gene.get_refseq(100_000_114, "T>A") == "115T>A"
    assert toy_gene.get_refseq(100_000_114, "T>A", from_atg=True) == "6T>A"
